package common;

import java.util.*;

/**
 * This is the <strike>Stack</strike>Heap Frame and primary data structure of Silver.
 * 
 * <p>Here be dragons, unfortunately. Don't modify at all without running it past me.
 * 
 * @author tedinski
 * @see Node
 */
public class DecoratedNode {

	// Information about this node
	protected final Node self;
	protected final DecoratedNode forwardParent;
	// parent may be actual parent, production with this as local, or production calling decorate
	protected final DecoratedNode parent;
	
	// caches
	protected final Object[] childrenValues;
	protected DecoratedNode forwardValue;
	protected final Map<String, Object> inheritedValues;
	// synthesized is also used for locals. (no name overlaps)
	protected final Map<String, Object> synthesizedValues;

	// Inherited attributes received from parent. (empty for forwarded-to nodes!)
	protected final Map<String, Lazy> inheritedAttributes;

	// This is necessary for 'let'.
	// TODO: modify 'let' to add its variables to locals normally
	// difficulty: have to add to the init stuff to Expr.  Expr lacks this.
	protected Map<String, Lazy> extraLocalAttributes;

	
	protected DecoratedNode(Node self, DecoratedNode parent,
			Map<String, Lazy> inhs, DecoratedNode forwardParent) {
		this.self = self;
		this.parent = parent;
		this.inheritedAttributes = inhs;
		this.forwardParent = forwardParent;
		
		// create caches
		if(self != null && self.getNumberOfChildren() > 0)
			this.childrenValues = new Object[self.getNumberOfChildren()];
		else
			this.childrenValues = null;
		if(self instanceof FunctionNode)
			this.inheritedValues = null;
		else
			this.inheritedValues = new TreeMap<String, Object>();
		this.synthesizedValues = new TreeMap<String, Object>();
		
		
		// STATS: Uncomment to enable statistics
		//Statistics.dnSpawn(self!=null?self.getClass():TopNode.class);
	}
	
	// STATS: Uncomment to enable statistics
	//@Override
	//protected void finalize() throws Throwable {
	//	Statistics.dnFinal(self!=null?self.getClass():TopNode.class);
	//}

	public DecoratedNode(Node self, DecoratedNode parent, Map<String, Lazy> inhs) {
		this(self,parent,inhs,null);
	}

	public DecoratedNode(Node self, DecoratedNode parent, DecoratedNode forwardParent) {
		this(self,parent,null,forwardParent);
	}

	public Node undecorate(){
		return self;
	}

	public Object child(int child){
		//System.err.println("TRACE: " + name + " demanding child: " + child);
		
		Object o = this.childrenValues[child]; 
		if(o == null) {
			o = self.getChild(child);
			// Thunk evaluation is now handled in Node, rather than here.
			if (o instanceof Node) {
				o = ((Node)o).decorate(this, self.getDefinedInheritedAttributes(child));
			}
			
			/*// "run time type checking" lots of false positives.
			if(o!= null)
			try {
				// This should never throw, since all Node children should have a static childTypes
				Class<?> cts[] = (Class<?>[])self.getClass().getField("childTypes").get(null);
				if(!cts[child].isInstance(o instanceof DecoratedNode ? ((DecoratedNode)o).self : o)) {
					
					// Functions take DecoratedNodes...
					if(cts[child] != common.DecoratedNode.class && cts[child] != common.Terminal.class) {
						System.out.println("RTTC: Node " + name + " getting child " + child +
										" failed.  Expected type: " + cts[child].getName() + 
										". Got type: " + o.getClass().getName());
						if(o instanceof DecoratedNode) {
							System.out.println("\t\t DC is of type: " + ((DecoratedNode)o).name);
						}
					}
				}
			} catch (Throwable t) {
				System.out.println("RTTC: Got exception? " + t);
			}*/
			
			// CACHE : probably should not comment out child caching?
			this.childrenValues[child] = o;
		}
		return o;
	}

	public Object local(String attribute) {
		//System.err.println("TRACE: " + name + " demanding local: " + attribute);
		
		Object o = this.synthesizedValues.get(attribute);
		if(o == null) {
			Lazy l = self.getLocal(attribute);
			if(l == null && this.extraLocalAttributes != null) {
				l = this.extraLocalAttributes.get(attribute);
			}
			if(l == null) {
				throw new RuntimeException("Local attribute '" + attribute + "' is not defined in production '" + self.getName() + "'");
			}
			try {
				o = l.eval(this);
				if (o instanceof Node) {
					o = ((Node)o).decorate(this, self.getDefinedInheritedAttributes(attribute));
				}
			} catch(Throwable t){
				throw new RuntimeException("Error evaluating local attribute '" + attribute + "' in production '" + self.getName() + "'", t);
			}
			
			// CACHE : comment out to disable caching for local attributes
			// not recommended due to IO objects (IOString, etc)
			this.synthesizedValues.put(attribute, o);
		}
		return o;
	}

	public Object synthesized(String attribute) {
		//System.err.println("TRACE: " + name + " demanding syn attribute: " + attribute);
		
		Object o = this.synthesizedValues.get(attribute);
		if(o == null) {
			Lazy l = self.getSynthesized(attribute);
			if(l != null) {
				try {
					o = l.eval(this);
				} catch(Throwable t) {
					throw new RuntimeException("Error evaluating synthesized attribute '" + attribute + "' in production '" + self.getName() + "'", t);
				}
			} else if(self.getForward() != null) {
				// TODO: tell forward not to cache the value since we will?
				try {
					o = forward().synthesized(attribute);
				} catch(Throwable t) {
					throw new RuntimeException("Error attempting to fetch from forward in production " + self.getName() + ".", t);
				}
			} else {
				throw new RuntimeException("Synthesized attribute '" + attribute + "' is not defined in production '" + self.getName() + "'");
			}
			
			// CACHE : comment out to disable caching for synthesized attributes
			this.synthesizedValues.put(attribute, o);
		}
		return o;
	}
	
	public DecoratedNode forward() {
		Lazy l = self.getForward();
		if(l == null)
			return null;
		if(this.forwardValue == null) {
			try {
				// CACHE : should not comment out forward caching !
				this.forwardValue = ((Node)l.eval(this)).decorate(parent, this);
			} catch(Throwable t) {
				throw new RuntimeException("Error evaluating forward node in production '" + self.getName() + "'", t);
			}
		}
		return this.forwardValue;
	}

	public Object inherited(String attribute) {
		//System.err.println("TRACE: " + name + " demanding inh attribute: " + attribute);
		
		Object o = this.inheritedValues.get(attribute);
		if(o == null) {
			Lazy l;
			// Note: inheritedAttributes is validly null here!  (forward nodes have no inherited attributes)
			if(this.inheritedAttributes == null || (l = this.inheritedAttributes.get(attribute)) == null) {
				if(forwardParent != null) {
					try {
						o = forwardParent.inheritedForwarded(attribute);
					} catch(Throwable t) {
						throw new RuntimeException("Inherited attribute '" + attribute + "' demanded by forward-to production'" + self.getName() +"'.", t);
					}
				} else {
					throw new RuntimeException("Inherited attribute '" + attribute + "' not provided to '" + self.getName() + "' by '" + parent.self.getName() + "'" + (forwardParent==null?"":" (forwardParent: '" + forwardParent.self.getName() + "')"));
				}
			} else {
				try {
					o = l.eval(this.parent);
				} catch(Throwable t) {
					throw new RuntimeException("Error evaluating inherited attribute '" + attribute + "' in production '" + self.getName() + "'", t);
				}
			}
			
			// CACHE : comment out to disable caching of inherited attributes
			// not recommended because it leads to a combinatoric explosion for environments
			// due to thunks keeping around references to previously computed environments that
			// we're recomputing
			this.inheritedValues.put(attribute, o);
		}
		return o;
	}
	
	// Called on a forwarding node by the forward target to retrieve inherited attributes.
	protected Object inheritedForwarded(String attribute) {
		//System.err.println("TRACE: " + name + " demanding FORWARDED inh attribute: " + attribute);
		
		Lazy l = self.getForwardInh(attribute);
		if(l == null) {
			return inherited(attribute);
		}
		try {
			// No need for caching here, it'll be cached by the inherited() that called us
			return l.eval(this);
		} catch(Throwable t) {
			throw new RuntimeException("Error evaluating inherited attribute '" + attribute + "' for forward production in production '" + self.getName() + "'", t);
		}
	}

}
