package common;

import java.util.*;

/**
 * Node represents undecorated nodes.  That is, we have children, but no inherited attributes, yet.
 * 
 * <p>This is subclassed by nonterminals (which are then subclassed by productions, which implement
 * the abstract methods).  Generally speaking, nothing is used on this class externally except the
 * various decorate methods, and by the DecoratedNode class.
 * 
 * @author tedinski
 * @see DecoratedNode
 */
public abstract class Node {
	
	protected final Object[] children;
	
	protected Node(Object[] children) {
		this.children = children;

		// STATS: Uncomment to enable statistics
		//Statistics.nSpawn(this.getClass());
	}
	
	// STATS: Uncomment to enable statistics
//	@Override
//	protected void finalize() throws Throwable {
//		Statistics.nFinal(this.getClass());
//	}

	public DecoratedNode decorate() {
		return decorate(TopNode.singleton);
	}

	public DecoratedNode decorate(DecoratedNode parent) {
		return decorate(parent, (Map<String, Lazy>)null);
	}

	public DecoratedNode decorate(DecoratedNode parent, Map<String, Lazy> inhs) {
		DecoratedNode d = new DecoratedNode(this, parent, inhs);
		return d;
	}

	public DecoratedNode decorate(DecoratedNode parent, DecoratedNode fwdParent) {
		DecoratedNode d = new DecoratedNode(this, parent, fwdParent);
		return d;
	}

	public Object getChild(int child) {
		Object o = children[child];
		if(o instanceof Thunk) {
			// We're doing thunk evaluation on the Node level, rather
			// than DecoratedNode.  This is more efficient, as there
			// may be more than one DecoratedNode per Node.
			// AND this eliminates all references to this thunk,
			// eliminating all references to that DecoratedNode and so on.
			o = ((Thunk)o).eval();
			children[child] = o;
		}
		return o;
	}
	
	public int getNumberOfChildren() {
		if(children == null)
			return 0;
		return children.length;
	}

	public abstract String getName();

	public abstract Lazy getForward();

	public abstract Lazy getForwardInh(String name);

	public abstract Lazy getSynthesized(String name);

	public abstract Map<String, Lazy> getDefinedInheritedAttributes(Object key);

	public abstract Lazy getLocal(String name);
}
